mod memory;
mod registers;

use crate::ppu::memory::PpuMemoryBus;
use crate::ppu::registers::{PpuControl, PpuInternalRegisters, PpuRegisters, PpuStatus};
use crate::ppu::registers::{
    PPU_REG_ADDR_ADDR, PPU_REG_ADDR_CTRL, PPU_REG_ADDR_DATA, PPU_REG_ADDR_MASK,
    PPU_REG_ADDR_OAMADDR, PPU_REG_ADDR_OAMDATA, PPU_REG_ADDR_OAMDMA, PPU_REG_ADDR_SCROLL,
    PPU_REG_ADDR_STATUS,
};
use crate::{Clock, MemoryBus, Observer};
use simplelog::{debug, info};

pub struct Ppu {
    registers: PpuRegisters,
    internal_registers: PpuInternalRegisters,
    memory_bus: PpuMemoryBus,
}

impl Ppu {
    pub fn new() -> Self {
        Self {
            registers: PpuRegisters {
                control: 0,
                mask: 0,
                status: 0,
                oam_addr: 0,
                oam_data: 0,
                scroll: 0,
                addr: 0,
                data: 0,
                oam_dma: 0,
            },
            internal_registers: PpuInternalRegisters {
                v: 0,
                t: 0,
                x: 0,
                w: false,
            },
            memory_bus: PpuMemoryBus::new(),
        }
    }

    pub fn get_background_pattern_table_addr(&self) -> u16 {
        let control = self.registers.control & PpuControl::BackgroundPatternTableAddress.bits();
        (control as u16) << 8
    }

    pub fn get_vram_addr_increment(&self) -> u16 {
        if self.registers.control & PpuControl::VramAdressIncrement.bits() > 0 {
            // Going down
            32
        } else {
            // Going across
            1
        }
    }

    pub fn read_register(&mut self, addr: u16) -> u8 {
        match addr {
            PPU_REG_ADDR_CTRL | PPU_REG_ADDR_MASK | PPU_REG_ADDR_OAMADDR | PPU_REG_ADDR_SCROLL
            | PPU_REG_ADDR_ADDR | PPU_REG_ADDR_OAMDMA => {
                assert!(false);
                0
            }
            PPU_REG_ADDR_STATUS => {
                self.internal_registers.w = false;
                let status = self.registers.status;

                // VBlank is cleared on read
                self.registers.status = !PpuStatus::VBlank.bits();

                status
            }
            PPU_REG_ADDR_OAMDATA => self.registers.oam_data,
            PPU_REG_ADDR_DATA => {
                // Access to VRAM memory is slow, so reading it a first time generally return the memory at the previous address.
                // So we get the data first, then update the register.
                let mut data = self.registers.data;
                self.registers.data = self.memory_bus.read_byte(self.internal_registers.v);

                if self.internal_registers.v > 0x3EFF {
                    // But the palette data is returned immediately
                    data = self.registers.data;
                }

                self.internal_registers.v += self.get_vram_addr_increment();
                data
            }
            _ => unreachable!(),
        }
    }

    pub fn write_register(&mut self, addr: u16, data: u8) {
        debug!("Write to PPU register: {:#04x} -> {:#02x}", addr, data);

        match addr {
            PPU_REG_ADDR_CTRL => {
                // TODO: Was broken, changed while loading palette
                // ppu_state.temp_ppu_addr = (ppu_state.temp_ppu_addr & 0xf3ff) | ((data & PPU_CTRL_BASE_NAMETABLE_ADDR) << 10);

                self.registers.control = data;

                // TODO: Trigger NMI
            }
            PPU_REG_ADDR_MASK => {
                self.registers.mask = data;
            }
            PPU_REG_ADDR_OAMADDR => {
                self.registers.oam_addr = data;
            }
            PPU_REG_ADDR_OAMDATA => {
                self.registers.oam_data = data;
            }
            PPU_REG_ADDR_SCROLL => {
                // TODO: Understand and fix with a game using scrolling
                if !self.internal_registers.w {
                    // Scroll X
                    self.internal_registers.t =
                        (self.internal_registers.t & 0xFFE0) | (data as u16 >> 3);
                    self.internal_registers.x = data & 0x7;
                } else {
                    // Scroll Y
                    self.internal_registers.t &= 0x0C1F;
                    self.internal_registers.t |= (data as u16 & 0xF8) << 2;
                    self.internal_registers.t |= (data as u16 & 0x7) << 12;
                    // TODO: ppu_state.y_scroll = data;
                }

                self.internal_registers.w = !self.internal_registers.w;
            }
            PPU_REG_ADDR_ADDR => {
                if !self.internal_registers.w {
                    self.internal_registers.t =
                        (self.internal_registers.t & 0xFF) | (data as u16 & 0x3F) << 8;
                } else {
                    self.internal_registers.t = (self.internal_registers.t & 0xFF00) | data as u16;
                    self.internal_registers.v = self.internal_registers.t;
                }

                self.registers.addr = data;
                self.internal_registers.w = !self.internal_registers.w;
            }
            PPU_REG_ADDR_DATA => {
                self.registers.data = data;
                let addr = self.internal_registers.v;
                self.memory_bus.write_byte(addr, data);
                self.internal_registers.v += self.get_vram_addr_increment();
            }
            PPU_REG_ADDR_OAMDMA => {
                self.registers.oam_dma = data;
            }
            PPU_REG_ADDR_STATUS => {
                assert!(false);
            }
            _ => unreachable!(),
        }
    }
}

impl Observer<u16, u8> for Ppu {
    fn notify_updated(&mut self, key: u16, value: u8) {
        if (key >= PPU_REG_ADDR_CTRL && key < PPU_REG_ADDR_DATA) || key == PPU_REG_ADDR_OAMDMA {
            self.write_register(key, value);
        }
    }
}

impl Clock for Ppu {
    fn cycle(&mut self) {
        // todo!()
    }
}
